import random
from telegram import Update, Bot
from ..responses.basic_responses import BaseResponse
from ..triggers.basic_triggers import BaseTrigger


class Handler():
    def __init__(self,
                 trigger_checker: BaseTrigger,
                 responder: BaseResponse,
                 probability: float = 1):
        self.trigger_checker = trigger_checker
        self.responder = responder

        if probability > 1 or probability <= 0:
            raise ValueError(f'Parameter \'probability\' should be in range (0, 1], '
                             f'but {probability} was given')

        self.probability_to_respond = probability

    def handle(self, bot: Bot, update: Update):
        trigger_result = self.trigger_checker.check_trigger(update)
        if trigger_result.should_respond:
            if random.random() <= self.probability_to_respond:
                self.responder.respond(bot, update, trigger_result.response_payload)



import abc
from typing import Optional
from telegram import Update
from .trigger_result import TriggerResult


class BaseTrigger(abc.ABC):
    @abc.abstractmethod
    def check_trigger(self, update: Update) -> TriggerResult:
        pass

    def __and__(self, other):
        return MergedTrigger(self, and_trigger=other)

    def __or__(self, other):
        return MergedTrigger(self, or_trigger=other)

    def __invert__(self):
        return InvertedTrigger(self)


class MergedTrigger(BaseTrigger):
    """Represents a trigger consisting of two other triggers.

    Args:
        base_trigger: Base trigger of the merged trigger
        and_trigger: Optional trigger to "and" with base_trigger.
            Mutually exclusive with or_trigger.
        or_trigger: Optional trigger to "or" with base_trigger.
            Mutually exclusive with and_trigger.

    """

    def __init__(self,
                 base_trigger: BaseTrigger,
                 and_trigger: BaseTrigger = None,
                 or_trigger: BaseTrigger = None):
        self.base_trigger = base_trigger
        self.and_trigger = and_trigger
        self.or_trigger = or_trigger

    def check_trigger(self, update: Update) -> Optional[TriggerResult]:
        if self.and_trigger:
            return self.check_trigger_and(update)

        if self.or_trigger:
            return self.check_trigger_or(update)

        return None

    def check_trigger_and(self, update: Update) -> TriggerResult:
        base_result = self.base_trigger.check_trigger(update)
        and_result = self.and_trigger.check_trigger(update)

        if base_result.should_respond and and_result.should_respond:
            merged_payload = self.merge_payloads(base_result, and_result)
            return TriggerResult(should_respond=True,
                                 response_payload=merged_payload)

        return TriggerResult(should_respond=False)

    def check_trigger_or(self, update: Update) -> TriggerResult:
        base_result = self.base_trigger.check_trigger(update)
        or_result = self.or_trigger.check_trigger(update)

        should_respond = False
        response_payload = {}
        if base_result.should_respond:
            should_respond = True
            if or_result.should_respond:
                response_payload = self.merge_payloads(base_result, or_result)
            else:
                response_payload = base_result.response_payload
        elif or_result.should_respond:
            should_respond = True
            response_payload = or_result.response_payload

        return TriggerResult(should_respond, response_payload)  # pytype: disable=wrong-arg-count

    @staticmethod
    def merge_payloads(first: TriggerResult, second: TriggerResult) -> dict:
        return {**first.response_payload, **second.response_payload}


class InvertedTrigger(BaseTrigger):
    """Represents a trigger that has been inverted.
    Args:
        original_trigger: The trigger to invert.
    """

    def __init__(self, original_trigger: BaseTrigger):
        self.original_trigger = original_trigger

    def check_trigger(self, update: Update) -> TriggerResult:
        original_result = self.original_trigger.check_trigger(update)
        return TriggerResult(should_respond=not original_result.should_respond)



from os import PathLike
from typing import Union
from .partial_ruamel_yaml import YAML
from .partial_ruamel_yaml import CommentedMap
from .rules_parsing_helper import RulesParsingHelper
from ..handlers.handler import Handler
from .trigger_response_params import TriggerParams, ResponseParams


class RulesParser:

    def __init__(self):
        self.yaml = YAML()
        self.global_triggers = {}
        self.global_responses = {}
        self.trigger_params = TriggerParams(global_elements=self.global_triggers)
        self.response_params = ResponseParams(global_elements=self.global_responses)

    def parse_globals(self, config: CommentedMap):
        RulesParsingHelper.add_globals(config, self.trigger_params)
        RulesParsingHelper.add_globals(config, self.response_params)

    def parse_single_rule(self, rule: CommentedMap):
        trigger = RulesParsingHelper.parse_rule_trigger_or_response(rule, self.trigger_params)
        response = RulesParsingHelper.parse_rule_trigger_or_response(rule, self.response_params)
        probability = rule['probability'] if 'probability' in rule else 1
        return Handler(trigger, response, probability=probability)

    def parse_file(self, file_path: Union[PathLike, str, bytes]):
        with open(file_path, 'r', encoding='utf-8') as stream:
            config = self.yaml.load(stream)

        self.parse_globals(config)

        return [self.parse_single_rule(rule) for rule in config['rules']]



import logging
from telegram.ext import Updater
from .configuration import token_file_path, rules_file_path
from .configuration.rules_parser import RulesParser
from .handlers.combined_handlers import CombinedConversationHandler
from .handlers.default_error_handler import handle_error


def start_bot():
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                        level=logging.INFO)

    with open(token_file_path(), 'r') as token_file:
        bot_token = token_file.read().strip()

    rule_parser = RulesParser()
    rule_handlers = rule_parser.parse_file(rules_file_path())
    conversation_handler = CombinedConversationHandler(rule_handlers)

    updater = Updater(bot_token)
    updater.dispatcher.add_handler(conversation_handler)
    updater.dispatcher.add_error_handler(handle_error)
    updater.start_polling()

    # Run the bot until you press Ctrl-C or the process receives SIGINT,
    # SIGTERM or SIGABRT. This should be used most of the time, since
    # start_polling() is non-blocking and will stop the bot gracefully.
    updater.idle()



from datetime import datetime
from typing import Optional
from telegram import Update
from .trigger_result import TriggerResult
from .basic_triggers import BaseTrigger


class EventStreakTrigger(BaseTrigger):
    """
    Event-streak trigger. This is used to trigger a rule in a case of a streak of events,
    for example when a certain phrase is written a few times in a row.

    This trigger gets a defined trigger as a "counting" event and optionally another one as a
    "resetting" event (to stop counting).
    It also gets the count of events to be considered a streak, and the timeout of the streak (the
    allowed time period between the first event and the last event in the streak.
    """

    def __init__(self,
                 streak_timeout_sec: int,
                 event_count: int,
                 counting_event_trigger: BaseTrigger,
                 resetting_event_trigger: Optional[BaseTrigger] = None):
        """
        Constructs the trigger.

        :param streak_timeout_sec: The allowed time period between the first event and the last \
        event in the streak
        :param event_count: The count of events to be considered a streak
        :param counting_event_trigger: A trigger to identify an event that counts towards the streak
        :param resetting_event_trigger: A trigger to identify an event that resets the streak
        """
        self.streak_timeout_sec = streak_timeout_sec
        self.event_count = event_count
        self.counting_event_trigger = counting_event_trigger

        if resetting_event_trigger is not None:
            self.resetting_event_trigger = resetting_event_trigger
        else:
            self.resetting_event_trigger = ~counting_event_trigger

        self.message_counter = 0
        self.streak_start_timestamp = 0

    def check_trigger(self, update: Update) -> TriggerResult:
        if self.counting_event_trigger.check_trigger(update).should_respond:
            now_timestamp = int(datetime.now().timestamp())

            if now_timestamp - self.streak_start_timestamp > self.streak_timeout_sec:
                self.reset_count()

            if self.message_counter == 0:
                self.streak_start_timestamp = int(datetime.now().timestamp())

            self.message_counter += 1

            if self.message_counter == self.event_count:
                self.reset_count()
                return TriggerResult(should_respond=True)

        elif self.resetting_event_trigger.check_trigger(update).should_respond:
            self.reset_count()

        return TriggerResult(should_respond=False)

    def reset_count(self):
        self.message_counter = 0



from telegram import Update
from telegram.ext import Filters, BaseFilter
from ..dict_enum import DictEnum
from .trigger_result import TriggerResult
from .basic_triggers import BaseTrigger
from ..users_helper import DEFAULT_USERS_HELPER


class FilterBasedTrigger(BaseTrigger):
    def __init__(self, message_filter: BaseFilter):
        self.filter = message_filter

    def check_trigger(self, update: Update) -> TriggerResult:
        return TriggerResult(should_respond=self.filter(update.message))


class _UserFilterBasedTrigger(FilterBasedTrigger):
    def __init__(self, nickname=None, user_id=None, username=None):
        if nickname is not None:
            super().__init__(Filters.user(DEFAULT_USERS_HELPER.get_user_id_by_nickname(nickname)))
        else:
            super().__init__(Filters.user(user_id, username))


class _ChatFilterBasedTrigger(FilterBasedTrigger):
    def __init__(self, chat_id=None, username=None):
        super().__init__(Filters.chat(chat_id, username))


class _LanguageFilterBasedTrigger(FilterBasedTrigger):
    def __init__(self, lang):
        super().__init__(Filters.language(lang))


class FilterTriggers(DictEnum):
    all = FilterBasedTrigger(Filters.all)
    text = FilterBasedTrigger(Filters.text)
    command = FilterBasedTrigger(Filters.command)
    reply = FilterBasedTrigger(Filters.reply)
    audio = FilterBasedTrigger(Filters.audio)
    document = FilterBasedTrigger(Filters.document)
    animation = FilterBasedTrigger(Filters.animation)
    photo = FilterBasedTrigger(Filters.photo)
    sticker = FilterBasedTrigger(Filters.sticker)
    video = FilterBasedTrigger(Filters.video)
    voice = FilterBasedTrigger(Filters.voice)
    video_note = FilterBasedTrigger(Filters.video_note)
    contact = FilterBasedTrigger(Filters.contact)
    location = FilterBasedTrigger(Filters.location)
    venue = FilterBasedTrigger(Filters.venue)
    status_update = FilterBasedTrigger(Filters.status_update)
    forwarded = FilterBasedTrigger(Filters.forwarded)
    game = FilterBasedTrigger(Filters.game)
    entity = FilterBasedTrigger(Filters.entity)  # Filters.entity is a class
    caption_entity = FilterBasedTrigger(Filters.caption_entity)  # Filters.caption_entity is a class
    private = FilterBasedTrigger(Filters.private)
    group = FilterBasedTrigger(Filters.group)
    invoice = FilterBasedTrigger(Filters.invoice)
    successful_payment = FilterBasedTrigger(Filters.successful_payment)
    passport_data = FilterBasedTrigger(Filters.passport_data)
    user = _UserFilterBasedTrigger
    chat = _ChatFilterBasedTrigger
    language = _LanguageFilterBasedTrigger



import re
from typing import List, Union
from telegram import Update
from .trigger_result import TriggerResult
from ..dict_enum import DictEnum
from .basic_triggers import BaseTrigger


class _RegExpTrigger(BaseTrigger):
    """
    Regular-expression-based trigger. This is used to trigger a rule when an incoming message
    matches the given pattern.
    """

    def __init__(self, pattern: str):
        """
        Constructs the trigger.

        :param pattern: The pattern to test the message's match with
        """
        self.pattern = pattern

    def check_trigger(self, update: Update) -> TriggerResult:
        if update.message.text is not None:
            text = update.message.text.replace('\n', ' ')
            match = re.match(self.pattern, text)
            if match is not None:
                return TriggerResult(should_respond=True,
                                     response_payload={'match': match.groups()})

        return TriggerResult(should_respond=False)


class _HasSubstringTrigger(_RegExpTrigger):
    """
    Substring trigger. This is used to trigger a rule when a certain substring  (or one of a list
    of substrings) exists in an incoming message.
    """

    def __init__(self, substring: Union[str, List[str]], exact: bool = False):
        """
        Constructs the trigger.

        :param substring: The substring/s to search in the message
        :param exact: Whether the exact substring should appear (as a whole word) in the message
        """

        if isinstance(substring, str):
            regexp_for_substring = substring
        elif isinstance(substring, list):
            if self.validate_strings_list(substring):
                regexp_for_substring = '|'.join(substring)
            else:
                raise TypeError('Parameter \'substring\' should be either a string or a list of '
                                'strings, but a list containing non-strings was given.')
        else:
            raise TypeError(f'Parameter \'substring\' should be either a string or a list of '
                            f'strings, but {type(substring)} was given')

        prefix = '(?:.* )?' if exact else '.*'
        postfix = '(?: .*)?' if exact else '.*'
        super().__init__(f'^{prefix}({regexp_for_substring}){postfix}$')

    @staticmethod
    def validate_strings_list(strings):
        return all(map(lambda element: isinstance(element, str), strings))


class _HasExactWordTrigger(_HasSubstringTrigger):
    """
    Word trigger - the same as substring trigger, but for exact words search. This is used to
    trigger a rule when a certain word (or one of a list of words) exists in an incoming message.
    """

    def __init__(self, word: Union[str, List[str]]):
        """
        Constructs the trigger.

        :param word: The word/s to search in the message
        """

        try:
            super().__init__(word, exact=True)
        except TypeError:
            raise TypeError(f'Parameter \'word\' should be either a string or a list of strings '
                            f'({type(word)} given)')


class TextTriggers(DictEnum):
    """Text-based triggers."""

    has_substring = _HasSubstringTrigger
    """A substring trigger. See more in :class:`_HasSubstringTrigger`."""

    has_exact_word = _HasExactWordTrigger
    """A word trigger. See more in :class:`_HasExactWordTrigger`."""

    regexp = _RegExpTrigger
    """A regexp-based trigger. See more in :class:`_RegExpTrigger`."""



from ..dict_enum import DictEnum
from .preset_responses import PresetResponses
from .match_responses import MatchResponses


class Responses(DictEnum):
    """A high-level class for all responses"""

    match = MatchResponses
    """Regexp-based responses. See more in :class:`MatchResponses`."""

    preset = PresetResponses
    """Preset responses. See more in :class:`PresetResponses`."""



from .partial_ruamel_yaml import CommentedMap
from .boolean_helper import BooleanHelper
from .trigger_response import TriggerResponse
from .trigger_response_params import TriggerResponseParams


class RulesParsingHelper:

    @staticmethod
    def add_globals(config: CommentedMap, params: TriggerResponseParams) -> None:
        """
        Adds globals (a generic name for "global triggers" and "global responses")
        from *config* to *params.globals*.

        Globals are triggers and responses which are defined generally in the rules file,
        as opposed to those defined inside a specific rule.

        :param config: The configuration subroot from which the function reads the
        triggers/responses.
        :param params: The trigger/response parameters whose globals should be updated.
        :return: None
        """
        if params.plural_key in config:
            parsed = params.parser.parse_many(config[params.plural_key], params.global_elements)
            params.global_elements.update(parsed)

    @staticmethod
    def parse_rule_trigger_or_response(rule: CommentedMap,
                                       params: TriggerResponseParams) -> TriggerResponse:

        return BooleanHelper.parse_subrule_as_trigger_or_response(rule[params.singular_key],
                                                                  params.global_elements,
                                                                  params.parser.parse_single)



from pathlib import Path


CONFIG_DIR = Path(Path.home(), '.gramhopper/')
TOKEN_FILE_NAME = 'token.txt'
RULES_FILE_NAME = 'rules.yml'
USERS_FILE_NAME = 'users.json'


def token_file_path():
    return Path(CONFIG_DIR, TOKEN_FILE_NAME)


def rules_file_path():
    return Path(CONFIG_DIR, RULES_FILE_NAME)


def users_file_path():
    return Path(CONFIG_DIR, USERS_FILE_NAME)



import abc
import random
from typing import Union, List
from telegram import Bot, Update, Document
from ..dict_enum import DictEnum
from .basic_responses import BaseResponse
from .response_helper import ResponseHelper


class _PresetTextResponse(BaseResponse):
    """
    A base class for preset responses. It is handling the response text building without
    handling the actual response action.
    """

    def __init__(self, preset_response: Union[str, List[str]]):
        """
        Constructs the response.

        :param preset_response: The preset response or list of responses
        """
        self.preset_responses = preset_response

    @abc.abstractmethod
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        pass

    def get_response_text(self):
        if isinstance(self.preset_responses, str):
            return self.preset_responses
        if isinstance(self.preset_responses, list):
            return random.choice(self.preset_responses)
        return None


class _PresetDocumentResponse(BaseResponse):
    """A preset response in which the response method is a document"""

    def __init__(self, preset_response: Union[str, Document]):
        """
        Constructs the response.

        :param preset_response: The preset document URL or document object
        """
        self.preset_response = preset_response

    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.document(bot, update, self.preset_response)


class _PresetMessageResponse(_PresetTextResponse):
    """A preset response in which the response method is a normal message"""

    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.message(bot, update, self.get_response_text())


class _PresetReplyResponse(_PresetTextResponse):
    """A preset response in which the response method is a reply to the triggering message"""
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.reply(bot, update, self.get_response_text())


class PresetResponses(DictEnum):
    """
    Preset responses.
    These responses use a preset response/s to respond with. If a list of responses is given,
    one of them will be chosen randomly for each response.
    """

    document = _PresetDocumentResponse
    """A preset **document** response. See more in :class:`_PresetDocumentResponse`."""

    message = _PresetMessageResponse
    """A preset **message** response. See more in :class:`_PresetMessageResponse`."""

    reply = _PresetReplyResponse
    """A preset **reply** response. See more in :class:`_PresetReplyResponse`."""



from telegram.ext import MessageHandler, Filters, ConversationHandler


REGULAR_STATE = 0


# A message handler that equally allows all of its child-handlers
# to handle any received update
class CombinedMessageHandler(MessageHandler):
    def __init__(self, handlers):
        self.handlers = handlers
        super().__init__(Filters.all, self.handle_all)

    def handle_all(self, bot, update):
        for handler in self.handlers:
            handler.handle(bot, update)
        return REGULAR_STATE


# A conversation handler that equally allows all of its child-handlers
# to handle any received update
class CombinedConversationHandler(ConversationHandler):
    def __init__(self, handlers):
        message_handlers = [CombinedMessageHandler(handlers)]
        super().__init__(entry_points=message_handlers,
                         states={REGULAR_STATE: message_handlers},
                         fallbacks=[])



import json
from os import PathLike
from typing import Union
from .configuration import users_file_path


class UsersHelper:

    def __init__(self, file_path: Union[PathLike, str, bytes] = None):
        if file_path is not None:
            self.file_path = file_path
        else:
            self.file_path = users_file_path()

        self._users = {}
        self.load_users()

    def load_users(self, force: bool = False):
        if force or not self._users:
            with open(self.file_path, 'r') as file:
                self._users = json.load(file)

    def get_user_id_by_nickname(self, nickname: str):
        return self._users[nickname]


DEFAULT_USERS_HELPER = UsersHelper()



import abc
import typing
from inspect import isclass
from .boolean_helper import BooleanHelper
from ..triggers.basic_triggers import BaseTrigger
from ..responses import Responses
from ..triggers import Triggers


class BaseParser(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def mapping_class():
        """ Returns the mapping class (`Triggers` or `Responses`)"""

    @classmethod
    def parse_single(cls, config, global_elements):  # pylint: disable=unused-argument
        config_copy = dict(config)
        if 'name' in config_copy:
            config_copy.pop('name')

        mapping_class = cls.mapping_class()
        element = mapping_class[config_copy.pop('type')]

        # Some triggers (most of them) are classes and some are instances (mostly filter triggers).
        # This allows both cases to be used.
        if isclass(element):
            return element(**config_copy)
        return element

    @classmethod
    def parse_many(cls, config, global_elements):
        return {
            element['name']: cls.parse_single(element, global_elements)
            for element
            in config
        }


class TriggerParser(BaseParser):
    @staticmethod
    def mapping_class():
        return Triggers

    @classmethod
    def parameters_to_parse_as_trigger(cls, config):
        trigger_class = Triggers[config['type']]
        parameters_type_hints = typing.get_type_hints(trigger_class.__init__)
        parameters_to_parse = []

        for parameter_name in config:

            # Meta-parameters to ignore
            if parameter_name in ['type', 'name']:
                continue

            # If this parameter has not type hint, we cannot check if it's a trigger or not
            if parameter_name not in parameters_type_hints:
                continue

            type_hint = parameters_type_hints[parameter_name]

            origin_type = getattr(type_hint, '__origin__', None)

            if origin_type is not None:
                if origin_type == typing.Union:
                    if BaseTrigger in type_hint.__args__:
                        parameters_to_parse.append(parameter_name)
                else:
                    raise NotImplementedError(f'Origin type {origin_type} '
                                              f'is currently not supported')
            elif isclass(type_hint) and issubclass(type_hint, BaseTrigger):
                parameters_to_parse.append(parameter_name)

        return parameters_to_parse

    @classmethod
    def parse_single(cls, config, global_elements):
        return BooleanHelper.parse_subrule_as_trigger_or_response(config,
                                                                  global_elements,
                                                                  cls.parse_single_recursively)

    @classmethod
    def parse_single_recursively(cls, config, global_elements):
        if isinstance(config, str):
            return global_elements[config]

        config_copy = dict(config)
        parameters_to_parse = cls.parameters_to_parse_as_trigger(config)
        config.pop('type')

        for parameter in parameters_to_parse:
            config_copy[parameter] = cls.parse_single(config[parameter], global_elements)

        return super().parse_single(config_copy, global_elements)


class ResponseParser(BaseParser):
    @staticmethod
    def mapping_class():
        return Responses



from typing import Any
from telegram import Bot, Update


class ResponseHelper:
    @staticmethod
    def reply(bot: Bot, update: Update, response: Any):
        bot.send_message(chat_id=update.effective_chat.id,
                         text=response,
                         reply_to_message_id=update.message.message_id)

    @staticmethod
    def message(bot: Bot, update: Update, response: Any):
        bot.send_message(chat_id=update.effective_chat.id, text=response)

    @staticmethod
    def document(bot: Bot, update: Update, response: Any):
        bot.send_document(update.effective_chat.id, response)



from ..dict_enum import DictEnum
from .filter_triggers import FilterTriggers
from .text_triggers import TextTriggers
from .event_streak_trigger import EventStreakTrigger


class Triggers(DictEnum):
    """A high-level class for all triggers"""

    event_streak = EventStreakTrigger
    """Event-streak trigger. See more in :class:`EventStreakTrigger`."""

    filter = FilterTriggers
    """Filter-based triggers. See more in :class:`FilterTriggers`."""

    text = TextTriggers
    """Text-based triggers. See more in :class:`TextTriggers`."""



from enum import Enum, EnumMeta


class DictEnumMeta(EnumMeta):
    """Metaclass for DictEnum"""

    def __getitem__(cls, key):
        key_path = key.split('.')
        item = super().__getitem__(key_path.pop(0)).value

        if key_path:
            return item['.'.join(key_path)]

        return item


class DictEnum(Enum, metaclass=DictEnumMeta):
    pass



from typing import Dict, Type
from dataclasses import dataclass
from .globals_dict import GlobalsDict
from .triggers_reponses_parsers import BaseParser, TriggerParser, ResponseParser
from ..triggers.basic_triggers import BaseTrigger
from ..responses.basic_responses import BaseResponse


@dataclass
class TriggerResponseParams:
    singular_key: str
    global_elements: GlobalsDict
    parser: Type[BaseParser]

    @property
    def plural_key(self):
        return self.singular_key + 's'


@dataclass
class TriggerParams(TriggerResponseParams):
    def __init__(self, global_elements: Dict[str, BaseTrigger]):
        super().__init__(singular_key='trigger',
                         global_elements=global_elements,
                         parser=TriggerParser)


@dataclass
class ResponseParams(TriggerResponseParams):
    def __init__(self, global_elements: Dict[str, BaseResponse]):
        super().__init__(singular_key='response',
                         global_elements=global_elements,
                         parser=ResponseParser)



from typing import Union, Callable
from boolean import boolean
from .partial_ruamel_yaml import CommentedMap
from .globals_dict import GlobalsDict
from .boolean_operators import OPERATOR_TYPE_TO_FUNCTION
from .trigger_response import TriggerResponse


class BooleanHelper:

    ParsingFunction = Callable[[Union[CommentedMap, str], GlobalsDict],
                               TriggerResponse]

    @staticmethod
    def evaluate_boolean_expression(expr: boolean.Expression,
                                    global_elements: GlobalsDict) -> TriggerResponse:
        # If the trigger/response here is just a name, look for it in the globals
        if isinstance(expr, boolean.Symbol):
            return global_elements[str(expr)]

        boolean_function = OPERATOR_TYPE_TO_FUNCTION[type(expr)]
        evaluated_args = [BooleanHelper.evaluate_boolean_expression(arg, global_elements)
                          for arg
                          in expr.args]
        return boolean_function(*evaluated_args)

    @staticmethod
    def parse_subrule_as_trigger_or_response(subrule: Union[CommentedMap, str],
                                             global_elements: GlobalsDict,
                                             parsing_func: ParsingFunction) -> TriggerResponse:
        if isinstance(subrule, str):
            algebra = boolean.BooleanAlgebra()
            parsed_expr = algebra.parse(subrule)
            return BooleanHelper.evaluate_boolean_expression(parsed_expr, global_elements)

        return parsing_func(subrule, global_elements)



import abc
from telegram import Bot, Update
from ..dict_enum import DictEnum
from .basic_responses import BaseResponse
from .response_helper import ResponseHelper


class _MatchTextResponse(BaseResponse):
    """
    A base class for regexp-based responses. It is handling the response text building without
    handling the actual response action.
    """

    def __init__(self, template: str):
        """
        Constructs the response.

        :param template: The template to use when building the response text
        """
        self.template = template

    @abc.abstractmethod
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        pass

    def build_response_text(self, response_payload: dict):
        return self.template.format(*response_payload['match'])


class _MatchMessageResponse(_MatchTextResponse):
    """A regexp-based response in which the response method is a normal message"""

    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.message(bot, update, self.build_response_text(response_payload))


class _MatchReplyResponse(_MatchTextResponse):
    """A regexp-based response in which the response method is a reply to the triggering message"""

    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.reply(bot, update, self.build_response_text(response_payload))


class MatchResponses(DictEnum):
    """
    Regexp-based responses.
    These responses use the regexp match result from the trigger, as well as the given template,
    to build the response text.
    """

    message = _MatchMessageResponse
    """A regexp-based **message** response. See more in :class:`_MatchMessageResponse`."""

    reply = _MatchReplyResponse
    """A regexp-based **reply** response. See more in :class:`_MatchReplyResponse`."""



"""
This file is a dirty hack to handle ruamel_yaml import problem
In certain distributions, the module name is ruamel_yaml, while in others it's ruamel.yaml.
As you can easily understand, that made us very sad :(
A fallback import wasn't enough to solve it since the same imports were done in 3 different files,
so we decided to create this file, which is a *very* partial accessor to ruamel_yaml.
"""

# Disabling pytype's import-error here because an actual import error would is caught here.
# Anyway, pylint itself throws an error if it can't find the imported module.
# pytype: disable=import-error
try:
    from ruamel_yaml import YAML
    from ruamel_yaml.comments import CommentedMap
except ImportError:
    from ruamel.yaml import YAML
    from ruamel.yaml.comments import CommentedMap


YAML = YAML
CommentedMap = CommentedMap  # pylint: disable=invalid-name



import logging
from telegram import Bot, Update, TelegramError


def handle_error(bot: Bot, update: Update, error: TelegramError):
    """Log Errors caused by Updates."""
    logging.error('Update "%s" caused error "%s" on bot "%s"', update, error, bot.name)



from setuptools import setup, find_namespace_packages

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='gramhopper',
    version='1.0.6',

    description='A bot platform for automatic responses based on various triggers',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",

    url='https://github.com/OrBin/gramhopper',

    author='Or Bin, Meir Halachmi',
    author_email='orbin50@gmail.com, meir.halachmi@gmail.com',
    install_requires=[
        'python_telegram_bot==11.1.0',
        'boolean.py==3.6',
        'ruamel_yaml~=0.15'
    ],

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Communications :: Chat',
    ],

    keywords='gramhopper telegram bot',

    packages=find_namespace_packages(include=['gramhopper', 'gramhopper.*']),

    entry_points={
        'console_scripts': [
            'gramhopper=gramhopper.bot:start_bot',
        ],
    },
)
